//
//  utils.h
//  mimircache
//
//  Created by Juncheng on 6/2/16.
//  Copyright © 2016 Juncheng. All rights reserved.
//

#ifndef UTILS_h
#define UTILS_h

#include <stdio.h>
#include <math.h> 
#include <glib.h> 



double get_log_base(guint64 max, guint64 expect_result);


#endif /* UTILS_h */
